#!/usr/bin/env python3



import binascii
import copy
import socket
import sys

import arguments
import decodeEMV

COMMANDS = [
#    '80A8000023832132204000000000000200000000000000082600000000000826130710003357a30a00',
#    '00B2010C00',
#    '00B2020C00',
#    '00B2011400',
#    '00B2021400',
#    '00B2031400',
#    '00B2041400',
#    '00B2051400',
#    '00B2011C00',
#    '00B2021C00',
]

def transceive(command, s):
    command = command.upper()
    print("Sending %s" % command)
    #print(s.getaddrinfo())
    cmd_bytes = bytearray.fromhex(command)
    #s.sendall(bytes([len(cmd_bytes)]))
    length = bytearray.fromhex("000000")
    length.extend([len(cmd_bytes)])
    s.sendall(length)
    print("Command length %s" %[len(cmd_bytes)])
    s.sendall(cmd_bytes)
    received_length = s.recv(1)
    length = int.from_bytes(received_length, byteorder='big')
    raw_response = s.recv(length)
    response = binascii.hexlify(bytearray(raw_response))
    fmt_response = response.decode('utf-8').upper()
    print("fmt response: " + fmt_response)
    return decodeEMV.parseEMV(fmt_response)

def pdol():
    # FIXME: parse these from App select response
    prod_life_cycle = "32204000"
    auth_amount = "000000000200" # £2.00
    #auth_amount = "000000005000" # £50.00
    unpred_num = "3357A30A"
    cur_code = "0826"
    pdol = prod_life_cycle + auth_amount + unpred_num + cur_code
    len_pdol_data = int(len(pdol) / 2)
    pdol_prefix = "%02x" % (len_pdol_data + 2) + "83" + "%02x" % (len_pdol_data)
    return pdol_prefix + pdol

def prep_get_proc_opts():
    return "80A80000"  + pdol() + "00"

def take_byte(s):
    idx = 2
    return (bytes.fromhex(s[:idx]), s[idx:])

def read_record_cmd(sfi, rec):
    # CLA
    ret = "00"
    # INS
    ret += "B2"
    # P1: Record number
    ret += rec.hex()
    # P2: Padded SFI
    # Set b3 to inform that P1 is a record number.
    new_sfi = int.to_bytes(sfi[0] | 0b100, 1, byteorder=sys.byteorder)
    ret += new_sfi.hex()
    # Le
    ret += "00"
    return ret

def parse_afl(afl):
    afl_len = len(afl) / 2
    # AFL entries must be 4 bytes in length each
    assert(float(afl_len) % 4 == 0)
    num_files = afl_len / 4

    ret = []
    while afl:
        # Get the details for this file.
        sfi, afl = take_byte(afl)
        current_rec, afl = take_byte(afl)
        last_rec, afl = take_byte(afl)
        num_dar, afl = take_byte(afl)
        last_idx = int.from_bytes(last_rec, sys.byteorder)
        rec_idx = int.from_bytes(current_rec, sys.byteorder)
        while rec_idx <= last_idx:
            # Read this record from the file.
            ret.append(read_record_cmd(sfi,
                                       rec_idx.to_bytes(1, sys.byteorder)))
            rec_idx += 1

    return ret

def generate_tvr():
    # Authorise all with 5 null bytes
    tac = "00" * 5
    # TODO: bitwise OR any IAC with our TAC to make TVR
    # set TVR = TAC for now.
    return tac

def generate_ac():
    # Generate AC?
    ret = "80AE"
    # P1: Offline auth/TC
    ret += "40"
    # P2
    ret += "00"

    # TODO: parse required data objects from CDOL1
    # TTQ: all 0 for now.
    data = "00"
    ret += data
    ret += "00"
    return ret

def fake_terminal():
    # SELECT ADF using AID.
    #transceive('00A404000e325041592E5359532E4444463031', s)
    transceive('00A404000E325041592E5359532E444446303100', s)

    # SELECT Visa app.
    transceive('00A4040007A000000004101000', s)
    #transceive('00A404000D325041592E5359532E4444463031', s)

    # GET processing options.
    transceive(prep_get_proc_opts(), s)

    # Get the commands to read each required file.
    # FIXME: parse AFL from Get Proc Opts response
    for cmd in parse_afl("10020500"):
        # Read the file.
        transceive(cmd, s)
    # qVSDC finishes here with card communication.

    # TODO: verify ICC public key certificate
    # TODO: static data authentication
    # TODO: dynamic data authentication
    # TODO: verify processing restrictions
    # TODO: perform CVM if CVM list exists...
    # TODO: perform Terminal Risk Management

    get_challenge = '0084000000'
    transceive(get_challenge, s)

    #transceive(generate_ac(), s)


if __name__ == '__main__':
    # Single socket only
    settings = arguments.parse(1)
    conn = settings.conns[0]

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:

        print(f"Connecting to host '{conn['host']}' on port '{conn['port']}'")
        s.connect((conn['host'], conn['port']))

        fake_terminal()
        print("Finished.")
